package com.holub.tools;

import java.util.EventObject;
import com.holub.tools.Subscriber;

/**	
 *	The Multicaster (modeled after the AWTEventMulticaster)
 *	provides an efficient way to manage relatively-short lists
 *	of "subscribers." Each Multcaster object can reference two
 *	"Subscriber" objects, one or both of which can be another
 *	Multicaster. The top-level Multicaster is passed a "publish"
 *	message, which it broadcasts (recursively) to both of the
 *	subscribers that it references.
 *	<p>
 *	The Multicaster is an "immutable" object, so you can't modify it.
 *	The "add()" method, for example, is passed two
 *	Multicasters and returns a third one that effectively references
 *	all the subscribers referenced by the original two. Any notifications
 *	that are in progress when the add() is executed will not be affected
 *	by the operation, however. It's perfectly acceptable for
 *	notifications to be performed on one thread while a second thread
 *	is adding or removing members from the Multcaster.
 *
 *	The order in which Subscribers are notified is undefined. (It is
 *	NOT necessarily the order of insertion.)
 *
 * @see java.awt.AWTEventMulticaster
 * @see Subscriber
 */

public class Multicaster implements Subscriber
{
	protected final Subscriber a, b;

    protected Multicaster(Subscriber a, Subscriber b)
    {	this.a = a;
    	this.b = b;
    }

	/******************************************************************
	 * Ask the subscribers of this multicaster to receive the
	 * <code>publication</code>. This is the "publish" operation, as
	 * seen from the perspective of a subscriber. (Remember, a multicaster
	 * is a list of subscribers.
	 *
	 * Note that the order of traversal should generally be considered
	 * undefined.
	 * However, if you really need to notify listeners in a known
	 * order, and you consistantly add nodes as follows:
	 * <pre>
	 *		subscription_list = Multicaster.add( subscription_list, new_node );
	 * </pre>
	 * (Where <code>subscription_list</code> is the head-of-list reference and
	 * <code>new_node</code> is the node you're adding), subscribers  will
	 * be notified in the order they were added. Removing nodes does not
	 * affect the order of notification. If you transpose the
	 * two arguments in the foregoing code:
	 * <pre>
	 *		subscription_list = Multicaster.add( new_node, subscription_list );
	 * </pre>
	 * subscribers will be notified in reverse order.
	 */

	public void receive( Object publication )
	{	a.receive( publication );
		b.receive( publication );
	}

	/******************************************************************
	 * Add a new subscriber to the list. The way that this call is used
	 * can impact the order in which subscribers are notified. (See
	 * {@link receive}.)
	 *
	 * @param a	Typically the head-of-list pointer.
	 * @param b	Typically the Subscriber you're adding to the list.
	 */
    public static Subscriber add(Subscriber a, Subscriber b)
    {	return  (a == null)  ? b :
				(b == null)  ? a : new Multicaster(a, b);
    }

	/******************************************************************
	 * Remove the indicated subscriber from the list
	 */
    public static Subscriber remove(Subscriber list, Subscriber remove_me)
    {
		if( list == remove_me || list == null  )
			return null;
		else if( !(list instanceof Multicaster) )
			return list;
		else
			return ((Multicaster)list).remove( remove_me );
    }

    private Subscriber remove(Subscriber remove_me)
    {
		if (remove_me == a)  return b;
		if (remove_me == b)  return a;

		Subscriber a2 = remove( a, remove_me );
		Subscriber b2 = remove( b, remove_me );

		return (a2 == a && b2 == b ) // it's not here
				? this
				: add(a2, b2)
				;
    }
	//================================================================
	public static class Test
	{	private static class Leaf implements Subscriber
		{	String s;
			public Leaf(String s){ this.s = s; }
			public void receive( Object publication )	
			{	System.out.print(s);
			}
		}

		public static void main( String[] args )
		{	Leaf a = new Leaf("A");
			Leaf b = new Leaf("B");
			Leaf c = new Leaf("C");
			Leaf d = new Leaf("D");
			Leaf e = new Leaf("E");

			Subscriber subscription_list = null;
			subscription_list = Multicaster.add( subscription_list, a );
			subscription_list = Multicaster.add( subscription_list, b );
			subscription_list = Multicaster.add( subscription_list, c );
			subscription_list = Multicaster.add( subscription_list, d );
			subscription_list = Multicaster.add( subscription_list, e );

			System.out.print("List is: ");
			subscription_list.receive( null );

			System.out.print("\nRemoving c: ");
			subscription_list = Multicaster.remove( subscription_list, c );
			subscription_list.receive( null );

			System.out.print("\nRemoving a: ");
			subscription_list = Multicaster.remove( subscription_list, a );
			subscription_list.receive( null );

			System.out.print("\nRemoving d: ");
			subscription_list = Multicaster.remove( subscription_list, d );
			subscription_list.receive( null );

			System.out.print("\nRemoving b: ");
			subscription_list = Multicaster.remove( subscription_list, b );
			subscription_list.receive( null );

			System.out.print("\nRemoving e: ");
			subscription_list = Multicaster.remove( subscription_list, e );

			if( subscription_list != null )
				System.out.println("Couldn't remove last node");

			subscription_list = Multicaster.add( a, subscription_list );
			subscription_list = Multicaster.add( b, subscription_list );
			subscription_list = Multicaster.add( c, subscription_list );
			subscription_list = Multicaster.add( d, subscription_list );
			subscription_list = Multicaster.add( e, subscription_list );

			System.out.print("\nShould be: EDCBA: " );
			subscription_list.receive( null );

			System.out.println();
		}
	}
}
